@@ -22,49 +22,53 @@ add features.</p>
 local fonts, logs, trackers, containers, resolvers = fonts, logs, trackers, containers, resolvers
 
 local next, type, tonumber, rawget = next, type, tonumber, rawget
-local match, gmatch, lower, gsub, strip, find = string.match, string.gmatch, string.lower, string.gsub, string.strip, string.find
-local char, byte, sub = string.char, string.byte, string.sub
+local match, gsub = string.match, string.gsub
 local abs = math.abs
-local bxor, rshift = bit32.bxor, bit32.rshift
 local P, S, R, Cmt, C, Ct, Cs, Carg = lpeg.P, lpeg.S, lpeg.R, lpeg.Cmt, lpeg.C, lpeg.Ct, lpeg.Cs, lpeg.Carg
 local lpegmatch, patterns = lpeg.match, lpeg.patterns
+local sortedhash = table.sortedhash
 
-local trace_features     = false  trackers.register("afm.features",   function(v) trace_features = v end)
-local trace_indexing     = false  trackers.register("afm.indexing",   function(v) trace_indexing = v end)
-local trace_loading      = false  trackers.register("afm.loading",    function(v) trace_loading  = v end)
-local trace_defining     = false  trackers.register("fonts.defining", function(v) trace_defining = v end)
+local trace_features      = false  trackers.register("afm.features",   function(v) trace_features = v end)
+local trace_indexing      = false  trackers.register("afm.indexing",   function(v) trace_indexing = v end)
+local trace_loading       = false  trackers.register("afm.loading",    function(v) trace_loading  = v end)
+local trace_defining      = false  trackers.register("fonts.defining", function(v) trace_defining = v end)
 
-local report_afm         = logs.reporter("fonts","afm loading")
+local report_afm          = logs.reporter("fonts","afm loading")
 
-local setmetatableindex  = table.setmetatableindex
-local derivetable        = table.derive
+local setmetatableindex   = table.setmetatableindex
+local derivetable         = table.derive
 
-local findbinfile        = resolvers.findbinfile
+local findbinfile         = resolvers.findbinfile
 
-local definers           = fonts.definers
-local readers            = fonts.readers
-local constructors       = fonts.constructors
+local privateoffset       = fonts.constructors and fonts.constructors.privateoffset or 0xF0000 -- 0x10FFFF
 
-local afm                = constructors.handlers.afm
-local pfb                = constructors.handlers.pfb
-local otf                = fonts.handlers.otf
+local definers            = fonts.definers
+local readers             = fonts.readers
+local constructors        = fonts.constructors
 
-local otfreaders         = otf.readers
-local otfenhancers       = otf.enhancers
+local afm                 = constructors.handlers.afm
+local pfb                 = constructors.handlers.pfb
+local otf                 = fonts.handlers.otf
 
-local afmfeatures        = constructors.features.afm
-local registerafmfeature = afmfeatures.register
+local otfreaders          = otf.readers
+local otfenhancers        = otf.enhancers
 
-afm.version              = 1.512 -- incrementing this number one up will force a re-cache
-afm.cache                = containers.define("fonts", "afm", afm.version, true)
-afm.autoprefixed         = true -- this will become false some day (catches texnansi-blabla.*)
+local afmfeatures         = constructors.features.afm
+local registerafmfeature  = afmfeatures.register
 
-afm.helpdata             = { }  -- set later on so no local for this
-afm.syncspace            = true -- when true, nicer stretch values
+local afmenhancers        = constructors.enhancers.afm
+local registerafmenhancer = afmenhancers.register
 
-local overloads          = fonts.mappings.overloads
+afm.version               = 1.513 -- incrementing this number one up will force a re-cache
+afm.cache                 = containers.define("fonts", "one", afm.version, true)
+afm.autoprefixed          = true -- this will become false some day (catches texnansi-blabla.*)
 
-local applyruntimefixes  = fonts.treatments and fonts.treatments.applyfixes
+afm.helpdata              = { }  -- set later on so no local for this
+afm.syncspace             = true -- when true, nicer stretch values
+
+local overloads           = fonts.mappings.overloads
+
+local applyruntimefixes   = fonts.treatments and fonts.treatments.applyfixes
 
 --[[ldx--
 <p>We cache files. Caching is taken care of in the loader. We cheat a bit by adding
@@ -76,42 +80,13 @@ fashion and later we transform it to sequences. Then we apply some methods also
 used in opentype fonts (like <t>tlig</t>).</p>
 --ldx]]--
 
-local enhancers = {
-    -- It's cleaner to implement them after we've seen what we are
-    -- dealing with.
-}
-
-local steps     = {
-    "unify names",
-    "add ligatures",
-    "add extra kerns",
-    "normalize features",
-    "fix names",
---  "add tounicode data",
-}
-
-local function applyenhancers(data,filename)
-    for i=1,#steps do
-        local step     = steps[i]
-        local enhancer = enhancers[step]
-        if enhancer then
-            if trace_loading then
-                report_afm("applying enhancer %a",step)
-            end
-            enhancer(data,filename)
-        else
-            report_afm("invalid enhancer %a",step)
-        end
-    end
-end
-
 function afm.load(filename)
     filename = resolvers.findfile(filename,'afm') or ""
     if filename ~= "" and not fonts.names.ignoredfile(filename) then
         local name = file.removesuffix(file.basename(filename))
         local data = containers.read(afm.cache,name)
         local attr = lfs.attributes(filename)
-        local size, time = attr.size or 0, attr.modification or 0
+        local size, time = attr and attr.size or 0, attr and attr.modification or 0
         --
         local pfbfile = file.replacesuffix(name,"pfb")
         local pfbname = resolvers.findfile(pfbfile,"pfb") or ""
@@ -128,7 +103,7 @@ function afm.load(filename)
             report_afm("reading %a",filename)
             data = afm.readers.loadfont(filename,pfbname)
             if data then
-                applyenhancers(data,filename)
+                afmenhancers.apply(data,filename)
              -- otfreaders.addunicodetable(data) -- only when not done yet
                 fonts.mappings.addtounicode(data,filename)
              -- otfreaders.extend(data)
@@ -161,13 +136,13 @@ end
 
 local uparser = fonts.mappings.makenameparser() -- each time
 
-enhancers["unify names"] = function(data, filename)
+local function enhance_unify_names(data, filename)
     local unicodevector = fonts.encodings.agl.unicodes -- loaded runtime in context
     local unicodes      = { }
     local names         = { }
-    local private       = constructors.privateoffset
+    local private       = data.private or privateoffset
     local descriptions  = data.descriptions
-    for name, blob in next, data.characters do
+    for name, blob in sortedhash(data.characters) do -- sorting is nicer for privates
         local code = unicodevector[name] -- or characters.name_to_unicode[name]
         if not code then
             code = lpegmatch(uparser,name)
@@ -205,19 +180,19 @@ enhancers["unify names"] = function(data, filename)
         end
     end
     data.characters = nil
+    data.private    = private
     local resources = data.resources
-    local filename = resources.filename or file.removesuffix(file.basename(filename))
+    local filename  = resources.filename or file.removesuffix(file.basename(filename))
     resources.filename = resolvers.unresolve(filename) -- no shortcut
     resources.unicodes = unicodes -- name to unicode
-    resources.marks = { } -- todo
- -- resources.names = names -- name to index
-    resources.private = private
+    resources.marks    = { } -- todo
+ -- resources.names    = names -- name to index
 end
 
 local everywhere = { ["*"] = { ["*"] = true } } -- or: { ["*"] = { "*" } }
 local noflags    = { false, false, false, false }
 
-enhancers["normalize features"] = function(data)
+local function enhance_normalize_features(data)
     local ligatures  = setmetatableindex("table")
     local kerns      = setmetatableindex("table")
     local extrakerns = setmetatableindex("table")
@@ -318,7 +293,7 @@ enhancers["normalize features"] = function(data)
     data.resources.sequences = sequences
 end
 
-enhancers["fix names"] = function(data)
+local function enhance_fix_names(data)
     for k, v in next, data.descriptions do
         local n = v.name
         local r = overloads[n]
@@ -365,14 +340,10 @@ local addthem = function(rawdata,ligatures)
     end
 end
 
-enhancers["add ligatures"] = function(rawdata)
+local function enhance_add_ligatures(rawdata)
     addthem(rawdata,afm.helpdata.ligatures)
 end
 
--- enhancers["add tex ligatures"] = function(rawdata)
---     addthem(rawdata,afm.helpdata.texligatures)
--- end
-
 --[[ldx--
 <p>We keep the extra kerns in separate kerning tables so that we can use
 them selectively.</p>
@@ -385,7 +356,7 @@ them selectively.</p>
 -- we don't use the character database. (Ok, we can have a context specific
 -- variant).
 
-enhancers["add extra kerns"] = function(rawdata) -- using shcodes is not robust here
+local function enhance_add_extra_kerns(rawdata) -- using shcodes is not robust here
     local descriptions = rawdata.descriptions
     local resources    = rawdata.resources
     local unicodes     = resources.unicodes
@@ -617,6 +588,7 @@ local function copytotfm(data)
         properties.fullname      = fullname
         properties.psname        = fullname
         properties.name          = filename or fullname or fontname
+        properties.private       = properties.private or data.private or privateoffset
         --
         if next(characters) then
             return {
@@ -752,18 +724,12 @@ end
 <p>We have the usual two modes and related features initializers and processors.</p>
 --ldx]]--
 
-local function setmode(tfmdata,value)
-    if value then
-        tfmdata.properties.mode = lower(value)
-    end
-end
-
 registerafmfeature {
     name         = "mode",
     description  = "mode",
     initializers = {
-        base = setmode,
-        node = setmode,
+        base = otf.modeinitializer,
+        node = otf.modeinitializer,
     }
 }
 
@@ -782,8 +748,6 @@ registerafmfeature {
 
 -- readers
 
-local check_tfm   = readers.check_tfm
-
 fonts.formats.afm = "type1"
 fonts.formats.pfb = "type1"
 
@@ -820,7 +784,8 @@ function readers.afm(specification,method)
             tfmdata = check_afm(specification,specification.name .. "." .. forced)
         end
         if not tfmdata then
-            method = method or definers.method or "afm or tfm"
+            local check_tfm = readers.check_tfm
+            method = (check_tfm and (method or definers.method or "afm or tfm")) or "afm"
             if method == "tfm" then
                 tfmdata = check_tfm(specification,specification.name)
             elseif method == "afm" then
@@ -855,3 +820,12 @@ function readers.pfb(specification,method) -- only called when forced
     swap("specification")
     return readers.afm(specification,method)
 end
+
+-- now we register them
+
+registerafmenhancer("unify names",          enhance_unify_names)
+registerafmenhancer("add ligatures",        enhance_add_ligatures)
+registerafmenhancer("add extra kerns",      enhance_add_extra_kerns)
+registerafmenhancer("normalize features",   enhance_normalize_features)
+registerafmenhancer("check extra features", otfenhancers.enhance)
+registerafmenhancer("fix names",            enhance_fix_names)
