@@ -2,14 +2,11 @@
 -----------------------------------------------------------------------
 --         FILE:  luaotfload-auxiliary.lua
 --  DESCRIPTION:  part of luaotfload
--- REQUIREMENTS:  luaotfload 2.6
+-- REQUIREMENTS:  luaotfload 2.91
 --       AUTHOR:  Khaled Hosny, Élie Roux, Philipp Gesang
 -----------------------------------------------------------------------
 --
 
---- this file addresses issue #24
---- https://github.com/lualatex/luaotfload/issues/24#
-
 luaotfload                  = luaotfload or { }
 local log                   = luaotfload.log
 local logreport             = log.report
@@ -42,7 +39,13 @@ local luaotfload_callbacks  = { }
 --- https://github.com/khaledhosny/luaotfload/issues/54
 
 local rewrite_fontname = function (tfmdata, specification)
-  tfmdata.name = [["]] .. specification .. [["]]
+  local format = tfmdata.format or tfmdata.properties.format
+  if stringfind (specification, " ") then
+    tfmdata.name = stringformat ("%q", specification)
+  else
+    --- other specs should parse just fine
+    tfmdata.name = specification
+  end
 end
 
 local rewriting = false
@@ -80,30 +83,33 @@ package.
 --doc]]--
 
 local set_sscale_dimens = function (fontdata)
-  local mathconstants = fontdata.MathConstants
-  local parameters    = fontdata.parameters
-  if mathconstants then
-    parameters[10] = mathconstants.ScriptPercentScaleDown or 70
-    parameters[11] = mathconstants.ScriptScriptPercentScaleDown or 50
-  end
-  return fontdata
+  local resources     = fontdata.resources      if not resources     then return end
+  local mathconstants = resources.MathConstants if not mathconstants then return end
+  local parameters    = fontdata.parameters     if not parameters    then return end
+  --- the default values below are complete crap
+  parameters [10] = mathconstants.ScriptPercentScaleDown       or 70
+  parameters [11] = mathconstants.ScriptScriptPercentScaleDown or 50
 end
 
 luaotfload_callbacks [#luaotfload_callbacks + 1] = {
   "patch_font", set_sscale_dimens, "set_sscale_dimens",
 }
 
+local default_units = 1000
+
 --- fontobj -> int
 local lookup_units = function (fontdata)
-  local metadata = fontdata.shared and fontdata.shared.rawdata.metadata
-  if metadata and metadata.units_per_em then
-    return metadata.units_per_em
-  elseif fontdata.parameters and fontdata.parameters.units then
-    return fontdata.parameters.units
-  elseif fontdata.units then --- v1.x
-    return fontdata.units
+  local units = fontdata.units
+  if units and units > 0 then return units end
+  local shared     = fontdata.shared    if not shared     then return default_units end
+  local rawdata    = shared.rawdata     if not rawdata    then return default_units end
+  local metadata   = rawdata.metadata   if not metadata   then return default_units end
+  local capheight  = metadata.capheight if not capheight  then return default_units end
+  local units      = metadata.units or fontdata.units
+  if not units or units == 0 then
+    return default_units
   end
-  return 1000
+  return units
 end
 
 --[[doc--
@@ -114,9 +120,9 @@ local patch_cambria_domh = function (fontdata)
   local mathconstants = fontdata.MathConstants
   if mathconstants and fontdata.psname == "CambriaMath" then
     --- my test Cambria has 2048
-    local units_per_em = fontdata.units_per_em or lookup_units(fontdata)
-    local sz           = fontdata.parameters.size or fontdata.size
-    local mh           = 2800 / units_per_em * sz
+    local units = fontdata.units or lookup_units(fontdata)
+    local sz    = fontdata.parameters.size or fontdata.size
+    local mh    = 2800 / units * sz
     if mathconstants.DisplayOperatorMinHeight < mh then
       mathconstants.DisplayOperatorMinHeight = mh
     end
@@ -169,35 +175,101 @@ Comment from fontspec:
 
 --doc]]--
 
-local set_capheight = function (fontdata)
-    local shared     = fontdata.shared
-    local parameters = fontdata.parameters
-    local capheight
-    if shared and shared.rawdata.metadata.pfminfo then
-      local units_per_em   = parameters.units
-      local size           = parameters.size
-      local os2_capheight  = shared.rawdata.metadata.pfminfo.os2_capheight
+local capheight_reference_chars      = { "X", "M", "Ж", "ξ", }
+local capheight_reference_codepoints do
+  local utfbyte = unicode.utf8.byte
+  capheight_reference_codepoints = { }
+  for i = 1, #capheight_reference_chars do
+    local chr = capheight_reference_chars [i]
+    capheight_reference_codepoints [i] = utfbyte (chr)
+  end
+end
 
-      if os2_capheight > 0 then
-        capheight = os2_capheight / units_per_em * size
-      else
-        local X8 = stringbyte"X"
-        if fontdata.characters[X8] then
-          capheight = fontdata.characters[X8].height
-        else
-          capheight = parameters.ascender / units_per_em * size
-        end
-      end
-    else
-      local X8 = stringbyte"X"
-      if fontdata.characters[X8] then
-        capheight = fontdata.characters[X8].height
-      end
+local determine_capheight = function (fontdata)
+  local parameters = fontdata.parameters if not parameters then return false end
+  local characters = fontdata.characters if not characters then return false end
+  --- Pretty simplistic but it does return *some* value for most fonts;
+  --- we could also refine the approach to return some kind of average
+  --- of all capital letters or a user-provided subset.
+  for i = 1, #capheight_reference_codepoints do
+    local refcp   = capheight_reference_codepoints [i]
+    local refchar = characters [refcp]
+    if refchar then
+      logreport ("both", 4, "aux",
+                 "picked height of character ‘%s’ (U+%d) as \\fontdimen8 \z
+                  candidate",
+                 capheight_reference_chars [i], refcp)
+      return refchar.height
     end
+  end
+  return false
+end
+
+local query_ascender = function (fontdata)
+  local parameters = fontdata.parameters if not parameters then return false end
+  local ascender   = parameters.ascender
+  if ascender then
+    return ascender --- pre-scaled
+  end
+
+  local shared     = fontdata.shared     if not shared     then return false end
+  local rawdata    = shared.rawdata      if not rawdata    then return false end
+  local metadata   = rawdata.metadata    if not metadata   then return false end
+  ascender         = metadata.ascender   if not ascender   then return false end
+  local size       = parameters.size     if not size       then return false end
+  local units = lookup_units (fontdata)
+  if not units or units == 0 then return false end
+  return ascender * size / units --- scaled
+end
+
+local query_capheight = function (fontdata)
+  local parameters = fontdata.parameters  if not parameters then return false end
+  local shared     = fontdata.shared      if not shared     then return false end
+  local rawdata    = shared.rawdata       if not rawdata    then return false end
+  local metadata   = rawdata.metadata     if not metadata   then return false end
+  local capheight  = metadata.capheight   if not capheight  then return false end
+  local size       = parameters.size      if not size       then return false end
+  local units = lookup_units (fontdata)
+  if not units or units == 0 then return false end
+  return capheight * size / units
+end
+
+local query_fontdimen8 = function (fontdata)
+  local parameters = fontdata.parameters if not parameters then return false end
+  local fontdimen8 = parameters [8]
+  if fontdimen8 then return fontdimen8 end
+  return false
+end
+
+local caphtfmt = function (ref, ht)
+  if not ht  then return "<none>"      end
+  if not ref then return tostring (ht) end
+  return stringformat ("%s(δ=%s)", ht, ht - ref)
+end
+
+local set_capheight = function (fontdata)
+    if not fontdata then
+      logreport ("both", 0, "aux",
+                 "error: set_capheight() received garbage")
+      return
+    end
+    local capheight_dimen8   = query_fontdimen8    (fontdata)
+    local capheight_alleged  = query_capheight     (fontdata)
+    local capheight_ascender = query_ascender      (fontdata)
+    local capheight_measured = determine_capheight (fontdata)
+    logreport ("term", 4, "aux",
+               "capht: param[8]=%s advertised=%s ascender=%s measured=%s",
+               tostring (capheight_dimen8),
+               caphtfmt (capheight_dimen8, capheight_alleged),
+               caphtfmt (capheight_dimen8, capheight_ascender),
+               caphtfmt (capheight_dimen8, capheight_measured))
+    if capheight_dimen8 then --- nothing to do
+      return
+    end
+
+    local capheight = capheight_alleged or capheight_ascender or capheight_measured
     if capheight then
-      --- is this legit? afaics there’s nothing else on the
-      --- array part of that table
-      fontdata.parameters[8] = capheight
+      fontdata.parameters [8] = capheight
     end
 end
 
@@ -250,20 +322,27 @@ end
 
 --- int -> string -> bool -> (int | false)
 local slot_of_name = function (font_id, glyphname, unsafe)
-  local fontdata = identifiers[font_id]
-  if fontdata then
-    local unicode = fontdata.resources.unicodes[glyphname]
-    if unicode then
-      if type(unicode) == "number" then
-        return unicode
-      else
-        return unicode[1] --- for multiple components
-      end
---  else
---    --- missing
+  if not font_id   or type (font_id)   ~= "number"
+  or not glyphname or type (glyphname) ~= "string"
+  then
+    logreport ("both", 0, "aux",
+               "invalid parameters to slot_of_name (%s, %s)",
+               tostring (font_id), tostring (glyphname))
+    return false
+  end
+
+  local tfmdata = identifiers [font_id]
+  if not tfmdata then return raw_slot_of_name (font_id, glyphname) end
+  local resources = tfmdata.resources  if not resources then return false end
+  local unicodes  = resources.unicodes if not unicodes  then return false end
+
+  local unicode = unicodes [glyphname]
+  if unicode then
+    if type (unicode) == "number" then
+      return unicode
+    else
+      return unicode [1] --- for multiple components
     end
-  elseif unsafe == true then -- for Robert
-    return raw_slot_of_name(font_id, glyphname)
   end
   return false
 end
@@ -285,11 +364,23 @@ local indices
 
 --- int -> (string | false)
 local name_of_slot = function (codepoint)
+  if not codepoint or type (codepoint) ~= "number" then
+    logreport ("both", 0, "aux",
+               "invalid parameters to name_of_slot (%s)",
+               tostring (codepoint))
+    return false
+  end
+
   if not indices then --- this will load the glyph list
     local unicodes = encodings.agl.unicodes
-    indices = table.swapped(unicodes)
+    if not unicodes or not next (unicodes)then
+      logreport ("both", 0, "aux",
+                 "name_of_slot: failed to load the AGL.")
+    end
+    indices = table.swapped (unicodes)
   end
-  local glyphname = indices[codepoint]
+
+  local glyphname = indices [codepoint]
   if glyphname then
     return glyphname
   end
@@ -298,137 +389,18 @@ end
 
 aux.name_of_slot      = name_of_slot
 
---[[doc--
-
-  In Context, characters.data is where the data from char-def.lua
-  resides. The file is huge (>3.7 MB as of 2013) and not part of the
-  isolated font loader. Nevertheless, we include a partial version
-  generated by the mkcharacters script that contains only the
-  a subset of the fields of each character defined.
-
-  Currently, these are (compare the mkcharacters script!)
-
-    · "direction"
-    · "mirror"
-    · "category"
-    · "textclass"
-
-  The directional information is required for packages like Simurgh [0]
-  to work correctly. In an early stage [1] it was necessary to load
-  further files from Context directly, including the full blown version
-  of char-def.  Since we have no use for most of the so imported
-  functionality, the required parts have been isolated and are now
-  instated along with luaotfload-characters.lua. We can extend the set
-  of imported features easily should it not be enough.
-
-  [0] https://github.com/persian-tex/simurgh
-  [1] http://tex.stackexchange.com/a/132301/14066
-
---doc]]--
-
-characters         = characters or { } --- should be created in basics-gen
-characters.data    = nil
-local chardef      = "luaotfload-characters"
-
-do
-  local setmetatableindex = function (t, f)
-    local mt = getmetatable (t)
-    if mt then
-      mt.__index = f
-    else
-      setmetatable (t, { __index = f })
-    end
-  end
-
-  --- there are some special tables for each field that provide access
-  --- to fields of the character table by means of a metatable
-
-  local mkcharspecial = function (characters, tablename, field)
-
-    local chardata = characters.data
-
-    if chardata then
-      local newspecial        = { }
-      characters [tablename]  = newspecial --> e.g. “characters.data.mirrors”
-
-      local idx = function (t, char)
-        local c = chardata [char]
-        if c then
-          local m = c [field] --> e.g. “mirror”
-          if m then
-            t [char] = m
-            return m
-          end
-        end
-        newspecial [char] = false
-        return char
-      end
-
-      setmetatableindex (newspecial, idx)
-    end
-
-  end
-
-  local mkcategories = function (characters) -- different from the others
-
-    local chardata = characters.data
-
-    setmetatable (characters, { __index = function (t, char)
-      if char then
-        local c = chardata [char]
-        c = c.category or char
-        t [char] = c
-        return c
-      end
-    end})
-
-  end
-
-  local load_failed = false
-  local chardata --> characters.data; loaded on demand
-
-  local load_chardef = function ()
-
-    logreport ("both", 1, "aux", "Loading character metadata from %s.", chardef)
-    chardata = dofile (kpse.find_file (chardef, "lua"))
-
-    if chardata == nil then
-      warning ("Could not load %s; continuing \z
-                with empty character table.",
-                chardef)
-      chardata    = { }
-      load_failed = true
-    end
-
-    characters      = { } --- nuke metatable
-    characters.data = chardata
-
-    --- institute some of the functionality from char-ini.lua
-
-    mkcharspecial (characters, "mirrors",     "mirror")
-    mkcharspecial (characters, "directions",  "direction")
-    mkcharspecial (characters, "textclasses", "textclass")
-    mkcategories  (characters)
-
-  end
-
-  local charindex = function (t, k)
-    if chardata == nil and load_failed ~= true then
-      load_chardef ()
-    end
-
-    return characters [k]
-  end
-
-  setmetatableindex (characters, charindex)
-
-end
 
 -----------------------------------------------------------------------
 ---                 features / scripts / languages
 -----------------------------------------------------------------------
 --- lots of arrowcode ahead
 
+local get_features = function (tfmdata)
+  local resources = tfmdata.resources  if not resources then return false end
+  local features  = resources.features if not features  then return false end
+  return features
+end
+
 --[[doc--
 This function, modeled after “check_script()” from fontspec, returns
 true if in the given font, the script “asked_script” is accounted for in at
@@ -437,13 +409,23 @@ least one feature.
 
 --- int -> string -> bool
 local provides_script = function (font_id, asked_script)
+  if not font_id      or type (font_id)      ~= "number"
+  or not asked_script or type (asked_script) ~= "string"
+  then
+    logreport ("both", 0, "aux",
+               "invalid parameters to provides_script(%s, %s)",
+               tostring (font_id), tostring (asked_script))
+    return false
+  end
   asked_script = stringlower(asked_script)
   if font_id and font_id > 0 then
-    local tfmdata  = identifiers[font_id] if not tfmdata  then return false end
-    local shared   = tfmdata.shared       if not shared   then return false end
-    local fontdata = shared.rawdata       if not fontdata then return false end
-    local fontname = fontdata.metadata.fontname
-    local features = fontdata.resources.features
+    local tfmdata = identifiers[font_id]
+    if not tfmdata then return false end
+    local features = get_features (tfmdata)
+    if features == false then
+      logreport ("log", 1, "aux", "font no %d lacks a features table", font_id)
+      return false
+    end
     for method, featuredata in next, features do
       --- where method: "gpos" | "gsub"
       for feature, data in next, featuredata do
@@ -474,14 +456,27 @@ feature.
 
 --- int -> string -> string -> bool
 local provides_language = function (font_id, asked_script, asked_language)
-  asked_script     = stringlower(asked_script)
-  asked_language   = stringlower(asked_language)
+  if not font_id        or type (font_id)        ~= "number"
+  or not asked_script   or type (asked_script)   ~= "string"
+  or not asked_language or type (asked_language) ~= "string"
+  then
+    logreport ("both", 0, "aux",
+               "invalid parameters to provides_language(%s, %s, %s)",
+               tostring (font_id),
+               tostring (asked_script),
+               tostring (asked_language))
+    return false
+  end
+  asked_script   = stringlower(asked_script)
+  asked_language = stringlower(asked_language)
   if font_id and font_id > 0 then
-    local tfmdata  = identifiers[font_id] if not tfmdata  then return false end
-    local shared   = tfmdata.shared       if not shared   then return false end
-    local fontdata = shared.rawdata       if not fontdata then return false end
-    local fontname = fontdata.metadata.fontname
-    local features = fontdata.resources.features
+    local tfmdata = identifiers[font_id]
+    if not tfmdata then return false end
+    local features = get_features (tfmdata)
+    if features == false then
+      logreport ("log", 1, "aux", "font no %d lacks a features table", font_id)
+      return false
+    end
     for method, featuredata in next, features do
       --- where method: "gpos" | "gsub"
       for feature, data in next, featuredata do
@@ -544,16 +539,29 @@ accounted for in the script with tag “asked_script” in feature
 --- int -> string -> string -> string -> bool
 local provides_feature = function (font_id,        asked_script,
                                    asked_language, asked_feature)
+  if not font_id        or type (font_id)        ~= "number"
+  or not asked_script   or type (asked_script)   ~= "string"
+  or not asked_language or type (asked_language) ~= "string"
+  or not asked_feature  or type (asked_feature)  ~= "string"
+  then
+    logreport ("both", 0, "aux",
+               "invalid parameters to provides_feature(%s, %s, %s, %s)",
+               tostring (font_id),        tostring (asked_script),
+               tostring (asked_language), tostring (asked_feature))
+    return false
+  end
   asked_script    = stringlower(asked_script)
   asked_language  = stringlower(asked_language)
   asked_feature   = lpegmatch(strip_garbage, asked_feature)
 
   if font_id and font_id > 0 then
-    local tfmdata  = identifiers[font_id] if not tfmdata  then return false end
-    local shared   = tfmdata.shared       if not shared   then return false end
-    local fontdata = shared.rawdata       if not fontdata then return false end
-    local features = fontdata.resources.features
-    local fontname = fontdata.metadata.fontname
+    local tfmdata  = identifiers[font_id]
+    if not tfmdata then return false end
+    local features = get_features (tfmdata)
+    if features == false then
+      logreport ("log", 1, "aux", "font no %d lacks a features table", font_id)
+      return false
+    end
     for method, featuredata in next, features do
       --- where method: "gpos" | "gsub"
       local feature = featuredata[asked_feature]
