@@ -3,12 +3,15 @@
 --         FILE:  luaotfload-init.lua
 --  DESCRIPTION:  Luaotfload font loader initialization
 -- REQUIREMENTS:  luatex v.0.80 or later; packages lualibs
---       AUTHOR:  Philipp Gesang (Phg), <phg@phi-gamma.net>
+--      VERSION:  2.91 2018-10-03
+--       AUTHOR:  Philipp Gesang (Phg), <phg@phi-gamma.net>, Marcel Krüger
 -----------------------------------------------------------------------
 --
 
+
 local setmetatable = setmetatable
 local kpselookup   = kpse.lookup
+local lfsisdir     = lfs.isdir
 
 --[[doc--
 
@@ -223,7 +226,7 @@ end
 
 --- below paths are relative to the texmf-context
 local ltx = "tex/generic/context/luatex"
-local ctx = "tex/context/base"
+local ctx = { "tex/context/base/mkiv", "tex/context/base" }
 
 local context_modules = {
 
@@ -237,31 +240,52 @@ local context_modules = {
   { false, "l-file"     },
   { false, "l-boolean"  },
   { false, "l-math"     },
+  { false, "l-unicode"  }, -- NEW UF 19.09.2018
   { false, "util-str"   },
+  { false, "util-fil"   },
 
   --- These constitute the fontloader proper.
   { ltx,   "luatex-basics-gen" },
   { ctx,   "data-con"          },
   { ltx,   "luatex-basics-nod" },
   { ctx,   "font-ini"          },
+  { ltx,   "luatex-fonts-mis"  }, -- NEW UF 19.09.2018
   { ctx,   "font-con"          },
   { ltx,   "luatex-fonts-enc"  },
   { ctx,   "font-cid"          },
-  { ltx,   "luatex-fonts-syn"  },
-  { ltx,   "luatex-fonts-tfm"  },
-  { ctx,   "font-oti"          },
-  { ctx,   "font-otf"          },
-  { ctx,   "font-otb"          },
-  { ctx,   "font-inj"          },
-  { ltx,   "luatex-fonts-ota"  },
-  { ctx,   "font-otn"          },
-  { ctx,   "font-otp"          },
-  { ltx,   "luatex-fonts-lua"  },
-  { ctx,   "font-def"          },
-  { ltx,   "luatex-fonts-def"  },
-  { ltx,   "luatex-fonts-ext"  },
-  { ltx,   "luatex-fonts-cbk"  },
   { ctx,   "font-map"          },
+--  { ltx,   "luatex-fonts-syn"  }, -- ignore??
+  { ctx,   "font-vfc"          }, -- NEW UF 19.09.2018
+  { ctx,   "font-oti"          },
+  { ctx,   "font-otr"          },
+  { ctx,   "font-ott"          }, 
+  { ctx,   "font-cff"          },
+  { ctx,   "font-ttf"          },
+  { ctx,   "font-dsp"          },
+  { ctx,   "font-oup"          },
+  { ctx,   "font-otl"          },
+  { ctx,   "font-oto"          },
+  { ctx,   "font-otj"          },
+  { ctx,   "font-ota"          },
+  { ctx,   "font-ots"          },
+  { ctx,   "font-osd"          },
+  { ctx,   "font-ocl"          },
+  { ctx,   "font-otc"          },
+  { ctx,   "font-onr"          },
+  { ctx,   "font-one"          },
+  { ctx,   "font-afk"          },
+  { ctx,   "font-tfm"          },
+  { ctx,   "font-lua"          },
+  { ctx,   "font-def"          },
+  { ltx,   "luatex-fonts-def"  },  
+  { ltx,   "luatex-fonts-ext"  },
+  { ctx,   "font-imp-tex"      },
+  { ctx,   "font-imp-ligatures"},
+  { ctx,   "font-imp-italics"  },
+  { ctx,   "font-imp-effects"  },
+  { ltx,   "luatex-fonts-lig"  },
+  { ltx,   "luatex-fonts-gbn"  },
+  
 
 } --[[context_modules]]
 
@@ -277,17 +301,37 @@ local load_context_modules = function (pth)
     local sub, spec = unpack (context_modules [i])
     if sub == false then
       ignore_module (spec)
-    elseif type (sub) == "string" then
-      if pth then
-        load_module (spec, file.join (pth, sub))
-      else
-        load_module (spec)
-      end
     else
-      logreport ("both", 0, "init",
-                 "Internal error, please report. \z
-                  This is not your fault.")
-      os.exit (-1)
+      local tsub = type (sub)
+      if not pth then
+        load_module (spec)
+      elseif tsub == "string" then
+        load_module (spec, file.join (pth, sub))
+      elseif tsub == "table" then
+        local pfx
+        local nsub = #sub
+        for j = 1, nsub do
+          local full = file.join (pth, sub [j])
+          if lfsisdir (full) then --- pick the first real one
+            pfx = full
+            break
+          end
+        end
+        if pfx then
+          load_module (spec, pfx)
+        else
+          logreport ("both", 0, "init",
+                     "None of the %d search paths for module %q exist; \z
+                      falling back to default path.",
+                     nsub, tostring (spec))
+          load_module (spec) --- maybe we’ll get by after all?
+        end
+      else
+        logreport ("both", 0, "init",
+                   "Internal error, please report. \z
+                    This is not your fault.")
+        os.exit (-1)
+      end
     end
   end
 
@@ -312,6 +356,137 @@ local init_adapt = function ()
 
 end --- [init_adapt]
 
+--[[doc--
+
+  In Context, characters.data is where the data from char-def.lua
+  resides. The file is huge (>4.4 MB as of 2016) and only a stripped
+  down version is part of the isolated font loader. Nevertheless, we
+  include an excerpt generated by the mkcharacters script that contains
+  a subset of the fields of each character defined and some extra
+  metadata.
+
+  Currently, these are (compare the mkcharacters script!)
+
+    · "direction"
+    · "mirror"
+    · "category"
+    · "textclass"
+
+  The directional information is required for packages like Simurgh [0]
+  to work correctly. In an early stage [1] it was necessary to load
+  further files from Context directly, including the full blown version
+  of char-def.  Since we have no use for most of the so imported
+  functionality, the required parts have been isolated and are now
+  instated along with luaotfload-characters.lua. We can extend the set
+  of imported features easily should it not be enough.
+
+  [0] https://github.com/persian-tex/simurgh
+  [1] http://tex.stackexchange.com/a/132301/14066
+
+--doc]]--
+
+characters         = characters or { } --- should be created in basics-gen
+characters.data    = nil
+local chardef      = "luaotfload-characters"
+
+do
+  local setmetatableindex = function (t, f)
+    local mt = getmetatable (t)
+    if mt then
+      mt.__index = f
+    else
+      setmetatable (t, { __index = f })
+    end
+  end
+
+  --- there are some special tables for each field that provide access
+  --- to fields of the character table by means of a metatable
+
+  local mkcharspecial = function (characters, tablename, field)
+
+    local chardata = characters.data
+
+    if chardata then
+      local newspecial        = { }
+      characters [tablename]  = newspecial --> e.g. “characters.data.mirrors”
+
+      local idx = function (t, char)
+        local c = chardata [char]
+        if c then
+          local m = c [field] --> e.g. “mirror”
+          if m then
+            t [char] = m
+            return m
+          end
+        end
+        newspecial [char] = false
+        return char
+      end
+
+      setmetatableindex (newspecial, idx)
+    end
+
+  end
+
+  local mkcategories = function (characters) -- different from the others
+
+    local chardata         = characters.data
+    local categories       = characters.categories or { }
+    characters.categories  = categories
+
+    setmetatable (categories, { __index = function (t, char)
+      if char then
+        local c = chardata [char]
+        c = c.category or char
+        t [char] = c
+        return c
+      end
+    end})
+
+  end
+
+  local load_failed = false
+  local chardata --> characters.data; loaded on demand
+
+  local load_chardef = function ()
+
+    logreport ("both", 1, "aux", "Loading character metadata from %s.", chardef)
+    chardata = dofile (kpse.find_file (chardef, "lua"))
+
+    if chardata == nil then
+      logreport ("both", 0, "aux",
+                 "Could not load %s; continuing with empty character table.",
+                 chardef)
+      chardata    = { }
+      load_failed = true
+    end
+
+    characters             = { } --- nuke metatable
+    characters.data        = chardata
+    characters.classifiers = chardata.classifiers
+    chardata.classifiers   = nil
+
+    --- institute some of the functionality from char-ini.lua
+
+    mkcharspecial (characters, "mirrors",     "mirror")
+    mkcharspecial (characters, "directions",  "direction")
+    mkcharspecial (characters, "textclasses", "textclass")
+    mkcategories  (characters)
+
+  end
+
+  local charindex = function (t, k)
+    if chardata == nil and load_failed ~= true then
+      load_chardef ()
+    end
+
+    return rawget (characters, k)
+  end
+
+  setmetatableindex (characters, charindex)
+
+end
+
 local init_main = function ()
 
   local load_fontloader_module = luaotfload.loaders.fontloader
@@ -322,9 +497,6 @@ local init_main = function ()
       Now that things are sorted out we can finally load the
       fontloader.
 
-      For less current distibutions we ship the code from TL 2014 that
-      should be compatible with Luatex 0.76.
-
   --doc]]--
 
   local fontloader = config.luaotfload and config.luaotfload.run.fontloader
@@ -362,37 +534,55 @@ local init_main = function ()
     ignore_module "l-boolean"
     ignore_module "l-math"
     ignore_module "util-str"
+    ignore_module "util-fil"
     ignore_module "luatex-basics-gen"
 
-    --- These constitute the fontloader proper.
     load_fontloader_module "data-con"
     load_fontloader_module "basics-nod"
     load_fontloader_module "font-ini"
+    load_fontloader_module "fonts-mis"
     load_fontloader_module "font-con"
     load_fontloader_module "fonts-enc"
     load_fontloader_module "font-cid"
     load_fontloader_module "font-map"
-    load_fontloader_module "fonts-syn"
-    load_fontloader_module "fonts-tfm"
-    load_fontloader_module "font-oti"
-    load_fontloader_module "font-otf"
-    load_fontloader_module "font-otb"
-    load_fontloader_module "font-inj"
-    load_fontloader_module "fonts-ota"
-    load_fontloader_module "font-otn"
-    load_fontloader_module "font-otp"   --> since 2013-04-23
-    load_fontloader_module "fonts-lua"
+--    load_fontloader_module "fonts-syn" -- ignore?
+    load_fontloader_module "font-vfc"
+    load_fontloader_module "font-otr"
+    load_fontloader_module "font-oti"    
+    load_fontloader_module "font-ott"
+    load_fontloader_module "font-cff"
+    load_fontloader_module "font-ttf"
+    load_fontloader_module "font-dsp"
+    load_fontloader_module "font-oup"
+    load_fontloader_module "font-otl"
+    load_fontloader_module "font-oto"
+    load_fontloader_module "font-otj"
+    load_fontloader_module "font-ota"
+    load_fontloader_module "font-ots"
+    load_fontloader_module "font-osd"
+    load_fontloader_module "font-ocl"
+    load_fontloader_module "font-otc"
+    load_fontloader_module "font-onr"
+    load_fontloader_module "font-one"
+    load_fontloader_module "font-afk"
+    load_fontloader_module "font-tfm"
+    load_fontloader_module "font-lua"
     load_fontloader_module "font-def"
     load_fontloader_module "fonts-def"
     load_fontloader_module "fonts-ext"
-    load_fontloader_module "fonts-cbk"
+    load_fontloader_module "font-imp-tex"
+    load_fontloader_module "font-imp-ligatures"
+    load_fontloader_module "font-imp-italics"
+    load_fontloader_module "font-imp-effects"
+    load_fontloader_module "fonts-lig"
+    load_fontloader_module "fonts-gbn"
 
   elseif fontloader == "context" then
     logreport ("log", 0, "init",
                "Loading Context modules in lookup path.")
     load_context_modules ()
 
-  elseif lfs.isdir (fontloader) then
+  elseif lfsisdir (fontloader) then
     logreport ("log", 0, "init",
                "Loading Context files under prefix “%s”.",
                fontloader)
@@ -411,18 +601,18 @@ local init_main = function ()
                fontloader, path)
     local _void = require (path)
 
-  elseif fontloader then
+  elseif kpselookup (("fontloader-%s.lua"):format(fontloader)) then
     logreport ("log", 0, "init",
                "Using predefined fontloader “%s”.",
                fontloader)
     load_fontloader_module (fontloader)
 
   else
-    logreport ("log", 0, "init",
+    logreport ("both", 0, "init",
                "No match for requested fontloader “%s”.",
                fontloader)
     fontloader = luaotfload.fontloader_package
-    logreport ("log", 0, "init",
+    logreport ("both", 0, "init",
                "Defaulting to predefined fontloader “%s”.",
                fontloader)
     load_fontloader_module (fontloader)
@@ -481,14 +671,20 @@ local init_post_install_callbacks = function ()
 
   --doc]]--
 
+  -- MK Pass current text direction to simple_font_handler
+  local handler = nodes.simple_font_handler
+  local callback = function(head, groupcode, _, _, direction)
+    return handler(head, groupcode, nil, nil, direction or tex.get'textdir')
+  end
   luatexbase.add_to_callback("pre_linebreak_filter",
-                             nodes.simple_font_handler,
+                             callback,
                              "luaotfload.node_processor",
                              1)
   luatexbase.add_to_callback("hpack_filter",
-                             nodes.simple_font_handler,
+                             callback,
                              "luaotfload.node_processor",
                              1)
+  -- /MK
 end
 
 local init_post_load_agl = function ()
