@@ -1,7 +1,7 @@
 if not modules then modules = { } end modules ['letterspace'] = {
-    version   = "2.5",
+    version   = "2.91",
     comment   = "companion to luaotfload-main.lua",
-    author    = "Hans Hagen, PRAGMA-ADE, Hasselt NL; adapted by Philipp Gesang",
+    author    = "Hans Hagen, PRAGMA-ADE, Hasselt NL; adapted by Philipp Gesang, Ulrike Fischer",
     copyright = "PRAGMA ADE / ConTeXt Development Team",
     license   = "see context related readme files"
 }
@@ -28,9 +28,6 @@ local setfield           = nodedirect.setfield
 local field_setter = function (name) return function (n, ...) setfield (n, name, ...) end end
 local field_getter = function (name) return function (n, ...) getfield (n, name, ...) end end
 
---- As of December 2014 the faster ``node.direct.*`` interface is
---- preferred.
-
 local getfont            = nodedirect.getfont
 local getid              = nodedirect.getid
 
@@ -64,12 +61,16 @@ local todirect           = nodedirect.tonut
 local tonode             = nodedirect.tonode
 
 local insert_node_before = nodedirect.insert_before
-local free_node          = nodedirect.free
+local free_node          = nodedirect.free -- may cause double free
+local free_node          = function (n)
+  logreport ("term", 5, "letterspace", "not calling free_node(%d)", n)
+  -- free_node (n)
+end
 local copy_node          = nodedirect.copy
 local new_node           = nodedirect.new
 
 local nodepool           = nodedirect.pool
-local new_kern           = nodepool.kern
+-- local new_kern           = nodepool.kern -- UF removed 2017-07-14
 
 local nodecodes          = nodes.nodecodes
 
@@ -235,16 +236,21 @@ end
 ---                 character kerning functionality
 ---=================================================================---
 
+-- UF changed 2017-07-14
+local nodedirectnew = node.direct.new
+
 local kern_injector = function (fillup, kern)
-  if fillup then
-    local g = new_glue(kern)
-    local s = getfield(g, "spec")
-    setfield(s, "stretch", kern)
-    setfield(s, "stretch_order", 1)
-    return g
-  end
-  return new_kern(kern)
+ if fillup then
+   local g = nodedirectnew("glue")
+   setfield(g, "stretch", kern)
+   setfield(g, "stretch_order", 1)
+   return g
+ end
+   local g = nodedirectnew("kern")
+   setfield(g,"kern",kern)
+   return g
 end
+-- /UF
 
 local kernable_skip = function (n)
   local st = getsubtype (n)
@@ -277,7 +283,7 @@ local kernfactors = { } --- fontid -> factor
 
 local kerncharacters
 kerncharacters = function (head)
-  local start, done   = head, false
+  local start         = head
   local lastfont      = nil
   local keepligature  = letterspace.keepligature --- function
   local keeptogether  = letterspace.keeptogether --- function
@@ -351,8 +357,7 @@ kerncharacters = function (head)
             end
             start = c
             setfield(s, "components", nil)
-            free_node(s)
-            done = true
+            free_node(s) --> double free with multipart components
             c = getfield (start, "components")
           end
         end
@@ -378,7 +383,6 @@ kerncharacters = function (head)
             local shrunk    = (getfield(spec,"shrink")  * newwd) / wd
             setfield(prev, "spec",
                      spec_injector(fillup, newwd, stretched, shrunk))
-            done = true
           end
 
         elseif pid == kern_code then
@@ -400,7 +404,6 @@ kerncharacters = function (head)
               local prev_kern = getfield(prev, "kern")
               prev_kern = prev_kern + quaddata[lastfont] * krn
               setfield (prev, "kern", prev_kern)
-              done = true
             end
           end
 
@@ -411,16 +414,23 @@ kerncharacters = function (head)
             if keeptogether and keeptogether(prev, start) then
               -- keep 'm
             elseif identifiers[lastfont] then
-              local kerns = chardata[lastfont] and chardata[lastfont][prevchar].kerns
-              local kern = kerns and kerns[lastchar] or 0
-              krn = kern + quaddata[lastfont]*krn -- here
-              insert_node_before(head,start,kern_injector(fillup,krn))
-              done = true
+              local lastfontchars = chardata[lastfont]
+              if lastfontchars then
+                local prevchardata  = lastfontchars[prevchar]
+                if not prevchardata then
+                  --- font doesn’t contain the glyph
+                else
+                  local kern = 0
+                  local kerns = prevchardata.kerns
+                  if kerns then kern = kerns[lastchar] or kern end
+                  krn = kern + quaddata[lastfont]*krn -- here
+                  insert_node_before(head,start,kern_injector(fillup,krn))
+                end
+              end
             end
           else
             krn = quaddata[lastfont]*krn -- here
             insert_node_before(head,start,kern_injector(fillup,krn))
-            done = true
           end
 
         elseif pid == disc_code then
@@ -477,10 +487,19 @@ kerncharacters = function (head)
               and getid(prv)   == glyph_code
               and getfont(prv) == lastfont
             then
+              local kern     = 0
               local prevchar = getchar(prv)
               local lastchar = getchar(start)
-              local kerns = chardata[lastfont] and chardata[lastfont][prevchar].kerns
-              local kern = kerns and kerns[lastchar] or 0
+              local lastfontchars = chardata[lastfont]
+              if lastfontchars then
+                local prevchardata = lastfontchars[prevchar]
+                if not prevchardata then
+                  --- font doesn’t contain the glyph
+                else
+                  local kerns = prevchardata.kerns
+                  if kerns then kern = kerns[lastchar] or kern end
+                end
+              end
               krn = kern + quaddata[lastfont]*krn -- here
             else
               krn = quaddata[lastfont]*krn -- here
@@ -496,7 +515,7 @@ kerncharacters = function (head)
       start = getnext(start)
     end
   end
-  return head, done
+  return head
 end
 
 ---=================================================================---
@@ -548,7 +567,7 @@ local enablefontkerning = function ( )
     logreport ("term", 5, "letterspace",
                "kerncharacters() invoked with node.direct interface \z
                (``%s`` -> ``%s``)", tostring (hd), tostring (direct_hd))
-    local direct_hd, _done = kerncharacters (direct_hd)
+    local direct_hd = kerncharacters (direct_hd)
     if not direct_hd then --- bad
       logreport ("both", 0, "letterspace",
                  "kerncharacters() failed to return a valid new head")
