@@ -2,17 +2,19 @@
 -------------------------------------------------------------------------------
 --         FILE:  luaotfload-parsers.lua
 --  DESCRIPTION:  various lpeg-based parsers used in Luaotfload
--- REQUIREMENTS:  Luaotfload >= 2.6
---       AUTHOR:  Philipp Gesang (Phg), <phg@phi-gamma.net>
+-- REQUIREMENTS:  Luaotfload >= 2.8
+--       AUTHOR:  Philipp Gesang (Phg), <phg@phi-gamma.net>, Marcel Krüger
 -------------------------------------------------------------------------------
 --
 
+
 local traversal_maxdepth  = 42 --- prevent stack overflows
 
 local rawset            = rawset
 
 local lpeg              = require "lpeg"
-local P, R, S           = lpeg.P, lpeg.R, lpeg.S
+local patterns          = lpeg.patterns
+local P, R, S, V        = lpeg.P, lpeg.R, lpeg.S, lpeg.V
 local lpegmatch         = lpeg.match
 local C, Cc, Cf         = lpeg.C, lpeg.Cc, lpeg.Cf
 local Cg, Cmt, Cs, Ct   = lpeg.Cg, lpeg.Cmt, lpeg.Cs, lpeg.Ct
@@ -45,6 +47,7 @@ local lfsisdir          = lfs.isdir
 ---                         COMMON PATTERNS
 -------------------------------------------------------------------------------
 
+local asterisk          = P"*"
 local dot               = P"."
 local colon             = P":"
 local semicolon         = P";"
@@ -56,6 +59,7 @@ local equals            = P"="
 local dash              = P"-"
 local gartenzaun        = P"#"
 local lbrk, rbrk        = P"[", P"]"
+local brackets          = lbrk + rbrk
 local squote            = P"'"
 local dquote            = P"\""
 
@@ -72,6 +76,8 @@ local digit             = R"09"
 local alpha             = R("az", "AZ")
 local anum              = alpha + digit
 local decimal           = digit^1 * (dot * digit^0)^-1
+local hexdigit          = R("09", "af", "AF")
+local hex               = P"0x" * hexdigit^1
 
 -------------------------------------------------------------------------------
 ---                                FONTCONFIG
@@ -233,6 +239,7 @@ end
 ---     -> (string -> fun option -> string list)
 ---     -> tab * tab * tab
 local read_fonts_conf_indeed
+-- MK Made basedir an explicit parameter to fix relative paths
 read_fonts_conf_indeed = function (depth,
                                    start,
                                    home,
@@ -241,7 +248,8 @@ read_fonts_conf_indeed = function (depth,
                                    acc,
                                    done,
                                    dirs_done,
-                                   find_files)
+                                   find_files,
+                                   basedir)
 
   logreport ("both", 4, "db",
              "Fontconfig scanner processing path %s.",
@@ -298,7 +306,7 @@ read_fonts_conf_indeed = function (depth,
       elseif --- if the path is relative, we make it absolute
         not ( lfsisfile(path) or lfsisdir(path) )
       then
-          path = filejoin(filedirname(start), path)
+          path = filejoin(basedir, path)
       end
       if  lfsisfile(path)
         and kpsereadable_file(path)
@@ -316,17 +324,22 @@ read_fonts_conf_indeed = function (depth,
                                        acc,
                                        done,
                                        dirs_done,
-                                       find_files)
+                                       find_files,
+                                       basedir)
         end
       elseif lfsisdir(path) then --- arrow code ahead
-        local config_files = find_files (path, conf_filter)
-        for _, filename in next, config_files do
-          if not done[filename] then
-            if done[path] then
+        if done[path] then
+          logreport ("log", 3, "db",
+                     "Skipping directory at %s, already included.", path)
+        else
+          done[path] = true
+          local config_files = find_files (path, conf_filter)
+          for _, filename in next, config_files do
+            if done[filename] then
               logreport ("log", 3, "db",
-                         "Skipping file at %s, already included.", path)
+                         "Skipping file at %s, already included.", filename)
             else
-              done[path] = true
+              done[filename] = true
               acc = read_fonts_conf_indeed(depth + 1,
                                            filename,
                                            home,
@@ -335,7 +348,8 @@ read_fonts_conf_indeed = function (depth,
                                            acc,
                                            done,
                                            dirs_done,
-                                           find_files)
+                                           find_files,
+                                           basedir)
             end
           end
         end
@@ -347,6 +361,7 @@ read_fonts_conf_indeed = function (depth,
   --inspect(done)
   return acc, done, dirs_done
 end --- read_fonts_conf_indeed()
+-- /MK
 
 --[[doc--
       read_fonts_conf() sets up an accumulator and two sets
@@ -371,6 +386,7 @@ local read_fonts_conf = function (path_list, find_files)
   local done      = { } ---> set:  files inspected
   local dirs_done = { } ---> set:  dirs in list
   for i=1, #path_list do --- we keep the state between files
+    -- MK Pass the base directory
     acc, done, dirs_done = read_fonts_conf_indeed(0,
                                                   path_list[i],
                                                   home,
@@ -379,7 +395,9 @@ local read_fonts_conf = function (path_list, find_files)
                                                   acc,
                                                   done,
                                                   dirs_done,
-                                                  find_files)
+                                                  find_files,
+                                                  filedirname(path_list[i]))
+    -- /MK
   end
   return acc
 end
@@ -560,9 +578,65 @@ local modifier          = slash * (other_modifier      --> ignore
                                  + garbage_modifier)   --> warn
 local modifier_list     = Cg(Ct(modifier^0), "modifiers")
 
+--- combining ---------------------------------------------------------
+---
+--- \font \three = combo: 1->42; 2->99,0xdeadbeef; 3->(1,U+2a-U+1337*42)
+---                       v  +~  v  +~ +~~~~~~~~~  v   v +~~~ +~~~~~ +~
+---                       |  |   |  |  |           |   | |    |      |
+---                       |  |   |  |  |           |   | |    |      |
+---        idx   :     ---+--÷---+--÷--÷-----------+   | +----+      |
+---                          |      |  |               | |           |
+---        id    :     ------+------+--÷---------------+ +-----------+
+---                                    |                 |
+---        chars :     ----------------+-----------------+
+---
+local comborowsep       = ws * semicolon * ws
+local combomapsep       = ws * P"->"     * ws
+local combodefsep       = ws * comma     * ws
+local comborangesep     = ws * asterisk  * ws
+
+local combohex          = hex     / tonumber
+local combouint         = digit^1 / tonumber
+local tonumber16        = function (s) return tonumber (s, 16) end
+local combouni          = P"U+" * (digit^1 / tonumber16)
+local combonum          = combohex + combouni + combouint
+local comborange        = Ct(combonum * dash * combonum) + combonum
+local combochars        = comborange * (comborangesep * comborange)^0
+local parenthesized     = function (p) return P"(" * ws * p * ws * P")" end
+
+local comboidxpat       = Cg(combouint, "idx")
+local comboidpat        = Cg(combouint, "id" )
+local combocharspat     = Cg(P"fallback" * Cc(true) + Ct(combochars^1), "chars")
+local comboidcharspat   = comboidpat * combodefsep * combocharspat
+
+local comboidx          = parenthesized (comboidxpat    ) + comboidxpat
+local comboid           = parenthesized (comboidpat     ) + comboidpat
+local comboidchars      = parenthesized (comboidcharspat) + comboidcharspat
+
+local combodef1         = Ct(comboidx * combomapsep * comboid) --> no chars
+local combodef          = Ct(comboidx * combomapsep * comboidchars)
+local combolist         = Ct(combodef1 * (comborowsep * combodef)^1)
+
+--- subfonts ----------------------------------------------------------
+--- This rule is present in the original parser. It sets the “sub”
+--- field of the specification which allows addressing a specific
+--- font inside a TTC container. Neither in Luatex-Fonts nor in
+--- Luaotfload is this documented, so we might as well silently drop
+--- it. However, as backward compatibility is one of our prime goals we
+--- just insert it here and leave it undocumented until someone cares
+--- to ask. (Note: afair subfonts are numbered, but this rule matches a
+--- string; I won’t mess with it though until someone reports a
+--- problem.)
+--- local subvalue   = P("(") * (C(P(1-S("()"))^1)/issub) * P(")") -- for Kim
+--- Note to self: subfonts apparently start at index 0. Tested with
+--- Cambria.ttc that includes “Cambria Math” at 0 and “Cambria” at 1.
+--- Other values cause luatex to segfault.
+local subfont           = P"(" * Cg((1 - S"()")^1, "sub") * P")"
+
 --- lookups -----------------------------------------------------------
 local fontname          = C((1-S":(/")^1)  --- like luatex-fonts
 local unsupported       = Cmt((1-S":(")^1, check_garbage)
+local combo             = P"combo:" * ws * Cg(combolist, "combo")
 local prefixed          = P"name:" * ws * Cg(fontname, "name")
 --- initially we intended file: to emulate the behavior of
 --- luatex-fonts, i.e. no paths allowed. after all, we do have XeTeX
@@ -572,15 +646,25 @@ local prefixed          = P"name:" * ws * Cg(fontname, "name")
 --- we’ll emit a warning.
                         + P"file:" * ws * Cg(unsupported, "path")
                         + P"file:" * ws * Cg(fontname, "file")
---- EXPERIMENTAL: kpse lookup
                         + P"kpse:" * ws * Cg(fontname, "kpse")
---- EXPERIMENTAL: custom lookup
                         + P"my:" * ws * Cg(fontname, "my")
 local unprefixed        = Cg(fontname, "anon")
-local path_lookup       = lbrk * Cg(C((1-rbrk)^1), "path") * rbrk
+--- Bracketed “path” lookups: These may contain any character except
+--- for unbalanced brackets. A backslash escapes any following
+--- character. Everything inside the outermost brackets is treated as
+--- part of the filename or path to look up. Subfonts may be specified
+--- in parens *after* the closing bracket. Note that this differs from
+--- Xetex whose syntax expects the subfont passed inside the brackets,
+--- separated by a colon.
+local path_escape       = backslash / "" * patterns.utf8char
+local path_content      = path_escape + (1 - brackets)
+local path_balanced     = { (path_content + V(2))^1
+                          , lbrk * V(1)^-1 * rbrk }
+local path_lookup       = lbrk * Cg(Cs(path_balanced), "path") * rbrk
+                        * subfont^-1
 
 --- features ----------------------------------------------------------
-local field_char        = anum + S"+-." --- sic!
+local field_char        = anum + S"+-.!?" --- sic!
 local field             = field_char^1
 --- assignments are “lhs=rhs”
 ---              or “+lhs=rhs” (Xetex-style)
@@ -603,22 +687,6 @@ local feature_list      = Cf(Ct""
                            * (featuresep * option^-1)^0
                            , rawset)
                         * featuresep^-1
-
---- other -------------------------------------------------------------
---- This rule is present in the original parser. It sets the “sub”
---- field of the specification which allows addressing a specific
---- font inside a TTC container. Neither in Luatex-Fonts nor in
---- Luaotfload is this documented, so we might as well silently drop
---- it. However, as backward compatibility is one of our prime goals we
---- just insert it here and leave it undocumented until someone cares
---- to ask. (Note: afair subfonts are numbered, but this rule matches a
---- string; I won’t mess with it though until someone reports a
---- problem.)
---- local subvalue   = P("(") * (C(P(1-S("()"))^1)/issub) * P(")") -- for Kim
---- Note to self: subfonts apparently start at index 0. Tested with
---- Cambria.ttc that includes “Cambria Math” at 0 and “Cambria” at 1.
---- Other values cause luatex to segfault.
-local subfont           = P"(" * Cg((1 - S"()")^1, "sub") * P")"
 --- top-level rules ---------------------------------------------------
 --- \font\foo=<specification>:<features>
 local features          = Cg(feature_list, "features")
@@ -626,6 +694,7 @@ local specification     = (prefixed + unprefixed)
                         * subfont^-1
                         * modifier_list^-1
 local font_request      = Ct(path_lookup   * (colon^-1 * features)^-1
+                           + combo --> TODO: feature list needed?
                            + specification * (colon    * features)^-1)
 
 --  lpeg.print(font_request)
@@ -678,10 +747,10 @@ local blank_line        = ws * eol
 local skip_line         = comment_line + blank_line
 local ini_id_char       = alpha + (dash / "_")
 local ini_id            = Cs(alpha * ini_id_char^0) / stringlower
-local ini_value_char    = (valid_escapes + (1 - newline - backslash - comment_char))
+local ini_value_char    = (valid_escapes + (1 - linebreak - backslash - comment_char))
 local ini_value         = (Cs (ini_value_char^0) / string.strip)
                         * (comment_char * (1 - eol)^0)^-1
-local ini_string_char   = (valid_escapes + (1 - newline - dquote - backslash))
+local ini_string_char   = (valid_escapes + (1 - linebreak - dquote - backslash))
 local ini_string        = dquote
                         * Cs (ini_string_char^0)
                         * dquote
